# -*- coding: utf-8 -*-
# imreg.py

# Copyright (c) 2014-?, Matěj Týč
# Copyright (c) 2011-2014, Christoph Gohlke
# Copyright (c) 2011-2014, The Regents of the University of California
# Produced at the Laboratory for Fluorescence Dynamics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
FFT based image registration. --- main functions
"""

from __future__ import division, print_function

import math

import numpy as np
import numpy.fft as fft
import scipy.ndimage.interpolation as ndii

import imreg_dft.utils as utils


__all__ = ['translation', 'similarity', 'transform_img',
           'transform_img_dict', 'imshow']


def _logpolar_filter(adft):
    shape = adft.shape
    yy = np.linspace(- np.pi / 2., np.pi / 2., shape[0])[:, np.newaxis]
    xx = np.linspace(- np.pi / 2., np.pi / 2., shape[1])[np.newaxis, :]
    # Supressing low spatial frequencies is a must when using log-polar
    # transform. The scale stuff is poorly reflected with low freqs.
    filt = 1.0 - np.cos(np.sqrt(yy ** 2 + xx ** 2))**2
    ret = adft * filt
    return ret


def _get_pcorr_shape(shape):
    ret = (int(max(shape) * 1.0),) * 2
    return ret


def _get_ang_scale(ims, bgval, exponent='inf', constraints=None):
    """
    Given two images, return their scale and angle difference.

    Args:
        ims (2-tuple-like of 2D ndarrays): The images
        bgval: We also pad here in the :func:`map_coordinates`
        exponent (float or 'inf'): The exponent stuff, see :func:`similarity`

    Returns:
        tuple: Scale, angle. Describes the relationship of the second image to
        the first one.
    """
    assert len(ims) == 2, \
        "Only two images are supported as input"
    shape = ims[0].shape

    adfts = [fft.fftshift(abs(fft.fft2(im))) for im in ims]
    adfts = [_logpolar_filter(adft) for adft in adfts]

    # High-pass filtering used to be here, but we have moved it to a higher
    # level interface

    pcorr_shape = _get_pcorr_shape(shape)
    log_base = _get_log_base(shape, pcorr_shape[1])
    stuffs = [_logpolar(adft, pcorr_shape, log_base, 0.0) for adft in adfts]

    if 0:
        import pylab as pyl
        pyl.figure(); pyl.imshow(ims[0]);
        pyl.figure(); pyl.imshow(ims[1]);
        pyl.show()

    (arg_ang, arg_rad), success = _phase_correlation(
        stuffs[0], stuffs[1],
        utils.argmax_angscale, log_base, exponent, constraints)

    angle = -np.pi * arg_ang / float(pcorr_shape[0])
    angle = np.rad2deg(angle)
    angle = utils.wrap_angle(angle, 360)
    scale = log_base ** arg_rad

    if not 0.5 < scale < 2:
        raise ValueError(
            "Images are not compatible. Scale change %g too big to be true."
            % scale)

    return 1.0 / scale, - angle


def _translation(im0, im2, filter_pcorr, odds=1, constraints=None):
    """
    Args:
        odds (float): The greater the odds are, the higher is the preferrence
            of the angle + 180 over the original angle. Odds of -1 are the same
            as inifinity.
    """
    angle = 0
    tvec, succ = translation(im0, im2, filter_pcorr, constraints)
    tvec2, succ2 = translation(im0, utils.rot180(im2),
                               filter_pcorr, constraints)
    if succ2 * odds > succ or odds == -1:
        tvec = tvec2
        succ = succ2
        angle += 180

    if 0:
        import pylab as pyl
        pyl.figure(); pyl.imshow(im0, cmap=pyl.cm.gray)
        pyl.show()
    return tvec, succ, angle


def _get_precision(shape, scale=1):
    pcorr_shape = _get_pcorr_shape(shape)
    log_base = _get_log_base(shape, pcorr_shape[1])
    # * 0.5 <= max deviation is half of the step
    # sccale: Scale deviation depends on the scale value
    Dscale = scale * (log_base - 1) * 0.5
    # angle: Angle deviation is constant
    Dangle = 180.0 / pcorr_shape[0] * 0.5
    return Dangle, Dscale


def similarity(im0, im1, numiter=1, order=3, constraints=None,
               filter_pcorr=0, exponent='inf'):
    """
    Return similarity transformed image im1 and transformation parameters.
    Transformation parameters are: isotropic scale factor, rotation angle (in
    degrees), and translation vector.

    A similarity transformation is an affine transformation with isotropic
    scale and without shear.

    Args:
        im0 (2D numpy array): The first (template) image
        im1 (2D numpy array): The second image
        numiter (int): How many times to iterate when determining scale and
            rotation
        order (int): Order of approximation (when doing transformations). 1 =
            linear, 3 = cubic etc.
        filter_pcorr (int): Radius of a spectrum filter for translation
            detection
        exponent (float or 'inf'): The exponent value used during processing.
            Refer to the docs for a thorough explanation. Generally, pass "inf"
            when feeling conservative. Otherwise, experiment, values below 5
            are not even supposed to work.

    Returns:
        dict: Contains following keys: ``scale``, ``angle``, ``tvec`` (Y, X),
        ``success`` and ``timg`` (the transformed image)

    .. note:: There are limitations

        * Scale change must be less than 2.
        * No subpixel precision (but you can use *resampling* to get
          around this).
    """
    shape = im0.shape
    if shape != im1.shape:
        raise ValueError("Images must have same shapes.")
    elif im0.ndim != 2:
        raise ValueError("Images must be 2-dimensional.")

    # We are going to iterate and precise scale and angle estimates
    scale = 1.0
    angle = 0.0
    im2 = im1

    if constraints is None:
        constraints = dict(angle=[0, None], scale=[1, None])

    # During iterations, we have to work with constraints too.
    # So we make the copy in order to leave the original intact
    constraints_dynamic = constraints.copy()
    constraints_dynamic["scale"] = list(constraints["scale"])
    constraints_dynamic["angle"] = list(constraints["angle"])

    bgval = utils.get_borderval(im1, 5)
    for ii in range(numiter):
        newscale, newangle = _get_ang_scale([im0, im2], bgval, exponent,
                                            constraints_dynamic)
        scale *= newscale
        angle += newangle

        constraints_dynamic["scale"][0] /= newscale
        constraints_dynamic["angle"][0] -= newangle

        im2 = transform_img(im1, scale, angle, bgval=bgval, order=order)

    # Here we look how is the turn-180
    target, stdev = constraints.get("angle", (0, None))
    odds = _get_odds(angle, target, stdev)

    # now we can use pcorr to guess the translation
    tvec, succ, angle2 = _translation(im0, im2, filter_pcorr, odds, constraints)

    # The log-polar transform may have got the angle wrong by 180 degrees.
    # The phase correlation can help us to correct that
    angle += angle2
    angle = utils.wrap_angle(angle, 360)

    # don't know what it does, but it alters the scale a little bit
    # scale = (im1.shape[1] - 1) / (int(im1.shape[1] / scale) - 1)

    Dangle, Dscale = _get_precision(shape, scale)

    res = dict(
        scale=scale,
        angle=angle,
        tvec=tvec,
        Dscale=Dscale,
        Dangle=Dangle,
        Dt=0.5,
        success=succ
    )

    im2 = transform_img_dict(im1, res, bgval, order)
    # Order of mask should be always 1 - higher values produce strange results.
    imask = transform_img_dict(np.ones_like(im1), res, 0, 1)
    # This removes some weird artifacts
    imask[imask > 0.8] = 1.0

    # Framing here = just blending the im2 with its BG according to the mask
    im2 = utils.frame_img(im2, imask, 10)

    res["timg"] = im2
    return res


def _get_odds(angle, target, stdev):
    """
    Args:

    Return:
        float: The greater the odds are, the higher is the preferrence
            of the angle + 180 over the original angle. Odds of -1 are the same
            as inifinity.
    """
    ret = 1
    if stdev is not None:
        diffs = [abs(utils.wrap_angle(ang, 360))
                 for ang in (target - angle, target - angle + 180)]
        odds0, odds1 = 0, 0
        if stdev > 0:
            odds0, odds1 = [np.exp(- diff ** 2 / stdev ** 2) for diff in diffs]
        if odds0 == 0 and odds1 > 0:
            # -1 is treated as infinity in _translation
            ret = -1
        elif stdev == 0 or (odds0 == 0 and odds1 == 0):
            ret = -1
            if diffs[0] < diffs[1]:
                ret = 0
        else:
            ret = odds1 / odds0
    return ret


def translation(im0, im1, filter_pcorr=0, constraints=None):
    """
    Return translation vector to register images.
    It tells how to translate the im1 to get im0.

    Args:
        im0 (2D numpy array): The first (template) image
        im1 (2D numpy array): The second image
        filter_pcorr (int): Radius of a spectrum filter for translation
            detection

    Returns:
        tuple: The translation vector and success number: ((Y, X), success)
    """
    ret, succ = _phase_correlation(
        im0, im1,
        utils.argmax_translation, filter_pcorr, constraints)
    return ret, succ


def _phase_correlation(im0, im1, callback=None, * args):
    """
    Computes phase correlation between im0 and im1

    Args:
        im0
        im1
        callback (function): Process the cross-power spectrum (i.e. choose
            coordinates of the best element, usually of the highest one).
            Defaults to :func:`utils.argmax2D`

    Returns:
        tuple: The translation vector (Y, X)
    """
    if callback is None:
        callback = utils.argmax2D
    f0, f1 = [fft.fft2(arr) for arr in (im0, im1)]
    # spectrum can be filtered, so we take precaution against dividing by 0
    eps = abs(f1).max() * 1e-15
    # cps == cross-power spectrum of im0 and im2
    cps = abs(fft.ifft2((f0 * f1.conjugate()) / (abs(f0) * abs(f1) + eps)))
    # scps = shifted cps
    scps = fft.fftshift(cps)

    if 0:
        import pylab as pyl
        pyl.figure(); pyl.imshow(im0);
        pyl.figure(); pyl.imshow(im1);
        pyl.figure(); pyl.imshow(scps);
        pyl.show()

    (t0, t1), success = callback(scps, * args)
    ret = np.array((t0, t1))

    # _compensate_fftshift is not appropriate here, this is OK.
    t0 -= f0.shape[0] // 2
    t1 -= f0.shape[1] // 2

    ret -= np.array(f0.shape, int) // 2
    return ret, success


def transform_img_dict(img, tdict, bgval=None, order=1, invert=False):
    """
    Wrapper of :func:`transform_img`, works well with the :func:`similarity`
    output.

    Args:
        img
        tdict (dictionary): Transformation dictionary --- supposed to contain
            keys "scale", "angle" and "tvec"
        bgval
        order
        invert (bool): Whether to perform inverse transformation --- doesn't
            work very well with the translation.

    Returns:
        :The same as :func:`transform_img`
    """
    scale = tdict["scale"]
    angle = tdict["angle"]
    tvec = np.array(tdict["tvec"])
    if invert:
        scale = 1.0 / scale
        angle *= -1
        tvec *= -1
    res = transform_img(img, scale, angle, tvec, bgval=bgval, order=order)
    return res


def transform_img(img, scale=1.0, angle=0.0, tvec=(0, 0), bgval=None, order=1):
    """
    Return translation vector to register images.

    Args:
        img (2D or 3D numpy array): What will be transformed.
            If a 3D array is passed, it is treated in a manner in which RGB
            images are supposed to be handled - i.e. assume that coordinates
            are (Y, X, channels).
        scale (float): The scale factor (scale > 1.0 means zooming in)
        angle (float): Degrees of rotation (clock-wise)
        tvec (2-tuple): Pixel translation vector, Y and X component.
        bgval (float): Shade of the background (filling during transformations)
            If None is passed, :func:`imreg_dft.utils.get_borderval` with
            radius of 5 is used to get it.
        order (int): Order of approximation (when doing transformations). 1 =
            linear, 3 = cubic etc. Linear works surprisingly well.

    Returns:
        The transformed img, may have another i.e. (bigger) shape than
            the source.
    """
    if img.ndim == 3:
        # A bloody painful special case of RGB images
        ret = np.empty_like(img)
        for idx in range(img.shape[2]):
            sli = (slice(None), slice(None), idx)
            ret[sli] = transform_img(img[sli], scale, angle, tvec,
                                     bgval, order)
        return ret

    if bgval is None:
        bgval = utils.get_borderval(img, 5)

    bigshape = np.array(img.shape) * 1.2
    bg = np.zeros(bigshape, img.dtype) + bgval

    dest0 = utils.embed_to(bg, img.copy())
    if scale != 1.0:
        dest0 = ndii.zoom(dest0, scale, order=order, cval=bgval)
    if angle != 0.0:
        dest0 = ndii.rotate(dest0, angle, order=order, cval=bgval)

    if tvec[0] != 0 or tvec[1] != 0:
        dest0 = ndii.shift(dest0, tvec, order=order, cval=bgval)

    bg = np.zeros_like(img) + bgval
    dest = utils.embed_to(bg, dest0)
    return dest


def similarity_matrix(scale, angle, vector):
    """
    Return homogeneous transformation matrix from similarity parameters.

    Transformation parameters are: isotropic scale factor, rotation angle (in
    degrees), and translation vector (of size 2).

    The order of transformations is: scale, rotate, translate.

    """
    raise NotImplementedError("We have no idea what this is supposed to do")
    m_scale = np.diag([scale, scale, 1.0])
    m_rot = np.identity(3)
    angle = math.radians(angle)
    m_rot[0, 0] = math.cos(angle)
    m_rot[1, 1] = math.cos(angle)
    m_rot[0, 1] = -math.sin(angle)
    m_rot[1, 0] = math.sin(angle)
    m_transl = np.identity(3)
    m_transl[:2, 2] = vector
    return np.dot(m_transl, np.dot(m_rot, m_scale))


def _get_log_base(shape, new_r):
    """
    Basically common functionality of :func:`_logpolar`
    and :func:`_get_ang_scale`

    This value can be considered fixed, if you want to mess with the logpolar
    transform, mess with the shape.
    """
    # The highest radius we have to accomodate is 'old_r',
    # However, we cut some parts out as only a thin part of the spectra has
    # these high frequencies
    old_r = shape[0] * 1.1
    # We are radius, so we divide the diameter by two.
    old_r /= 2.0
    # we have at most 'new_r' of space.
    # the base is chosen so that 'new_r' = log_'base'('old_r')
    log_base = np.exp(np.log(old_r) / (new_r))
    return log_base


def _logpolar(image, shape, log_base, bgval=None):
    """
    Return log-polar transformed image and log base.
    Takes into account anisotropicity of the freq spectrum of rectangular images

    Args:
        image: The image to be transformed
        shape: Shape of the transformed image
        log_base: Parameter of the transformation, convoluted with
            :func:`_get_log_base`

    Returns:
        The transformed image
    """
    if bgval is None:
        bgval = utils.get_borderval(image, 5)
    imshape = np.array(image.shape)
    center = imshape[0] / 2.0, imshape[1] / 2.0
    # 0 .. pi = only half of the spectrum is used
    theta = utils._get_angles(shape)
    radius_x = utils._get_scales(shape, log_base)
    radius_y = radius_x.copy()
    ellipse_coef = imshape[0] / float(imshape[1])
    # We have to acknowledge that the frequency spectrum can be deformed
    # if the image aspect ratio is not 1.0
    # The image is x-thin, so we acknowledge that the frequency spectra
    # scale in x is shrunk.
    radius_x /= ellipse_coef

    y = radius_y * np.sin(theta) + center[0]
    x = radius_x * np.cos(theta) + center[1]
    output = np.empty_like(y)
    ndii.map_coordinates(image, [y, x], output=output, order=3,
                         mode="constant", cval=bgval)
    """
    import pylab as pyl
    pyl.figure(); pyl.imshow(output);
    pyl.show()
    """
    return output


def imshow(im0, im1, im2, cmap=None, fig=None, **kwargs):
    """
    Plot images using matplotlib.
    Opens a new figure with four subplots:

    ::

      +---------------------+---------------------+
      |                     |                     |
      |   <template image>  |       <image>       |
      |                     |                     |
      +---------------------+---------------------+
      | <difference between |                     |
      |   template and the  | <transformed image> |
      |  transformed image> |                     |
      +---------------------+---------------------+

    Args:
        im0 (np.ndarray): The template image
        im1: The ``image``
        im2: The transformed ``image`` --- it is supposed to match the template
        cmap (optional): colormap
        fig (optional): The figure you would like to have this plotted on

    Returns:
        matplotlib figure: The figure with subplots
    """
    from matplotlib import pyplot

    if fig is None:
        fig = pyplot.figure()
    if cmap is None:
        cmap = 'coolwarm'
    # We do the difference between the template and the result now
    # To increase the contrast of the difference, we norm images according
    # to their near-maximums
    norm = np.percentile(im2, 99.5) / np.percentile(im0, 99.5)
    im3 = abs(im2 - im0 * norm)
    pl0 = fig.add_subplot(221)
    pl0.imshow(im0, cmap, **kwargs)
    pl0.grid()
    share = dict(sharex=pl0, sharey=pl0)
    pl = fig.add_subplot(222, ** share)
    pl.imshow(im1, cmap, **kwargs)
    pl.grid()
    pl = fig.add_subplot(223, ** share)
    pl.imshow(im3, cmap, **kwargs)
    pl.grid()
    pl = fig.add_subplot(224, ** share)
    pl.imshow(im2, cmap, **kwargs)
    pl.grid()
    return fig



import argparse as ap

import scipy.io as io
import scipy.misc as misc


def parse():
    parser = ap.ArgumentParser()
    parser.add_argument('imgfile')
    parser.add_argument('outfile')
    parser.add_argument('--var', default="img")
    parser.add_argument('--dummy-vars', default="", metavar="NAME1,NAME2,...")
    args = parser.parse_args()
    return args


def main():
    args = parse()
    img = misc.imread(args.imgfile)
    tosave = {}
    for dvar in args.dummy_vars.split(","):
        if len(dvar) == 0:
            continue
        tosave[dvar] = 0
    tosave[args.var] = img
    io.savemat(args.outfile, tosave)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# cli.py

# Copyright (c) 2014-?, Matěj Týč
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
FFT based image registration. --- CLI frontend
"""

import sys
import argparse as ap

import imreg_dft as ird
import imreg_dft.loader as loader


def assure_constraint(possible_constraints):
    pass


def _constraints(what):
    BOUNDS = dict(
        angle=(-180, 180),
        scale=(0.5, 2.0),
    )

    def constraint(string):
        components = string.split(",")
        if not (0 < len(components) <= 2):
            raise ap.ArgumentTypeError(
                "We accept at most %d (but at least 1) comma-delimited numbers,"
                " you have passed us %d" % (len(components), 2))
        try:
            mean = float(components[0])
        except Exception:
            raise ap.ArgumentTypeError(
                "The %s value must be a float number, got '%s'."
                % (what, components[0]))
        if what in BOUNDS:
            lo, hi = BOUNDS[what]
            if not lo <= mean <= hi:
                raise ap.ArgumentTypeError(
                    "The %s value must be a number between %g and %g, got %g."
                    % (lo, hi, mean))
        if len(components) == 2:
            std = components[1]
            if len(std) == 0:
                std = None
            else:
                try:
                    std = float(std)
                except Exception:
                    raise ap.ArgumentTypeError(
                        "The %s standard deviation spec must be either"
                        "either a float number or nothing, got '%s'."
                        % (what, std))
        ret = (mean, std)
        return ret
    return constraint


def _float_tuple(string):
    """
    Support function for parsing string of two floats delimited by a comma.
    """
    vals = string.split(",")
    if len(vals) != 2:
        raise ap.ArgumentTypeError(
            "'%s' are not two values delimited by comma" % string)
    try:
        vals = [float(val) for val in vals]
    except ValueError:
        raise ap.ArgumentTypeError("%s are not two float values" % vals)
    return vals


def _exponent(string):
    """
    Converts the passed string to a float or "inf"
    """
    if string == 'inf':
        return string
    try:
        ret = float(string)
    except:
        raise ap.ArgumentTypeError(
            "'%s' should be either 'inf' or a float value" % string)
    return ret


def outmsg(msg):
    """
    Support function for checking of validity of the output format string.
    A test interpolation is performed and exceptions handled.
    """
    fake_data = dict(scale=1.0, angle=2.0, tx=2, ty=2,
                     Dscale=0.1, Dangle=0.2, Dt=0.5, success=0.99)
    tpl = "The string '%s' is not a good format string"
    try:
        msg % fake_data
    except KeyError as exc:
        raise ap.ArgumentTypeError(
            (tpl + ". The correct string "
             "has to contain at most %s, but this one also contains an invalid"
             " value '%s'.") % (msg, fake_data.keys(), exc.args[0]))
    except Exception as exc:
        raise ap.ArgumentTypeError(
            (tpl + " - %s") % (msg, exc.message))
    return msg


def create_parser():
    parser = ap.ArgumentParser()
    parser.add_argument('template')
    parser.add_argument('image')
    parser.add_argument('--show', action="store_true", default=False,
                        help="Whether to show registration result")
    parser.add_argument('--lowpass', type=_float_tuple,
                        default=None, metavar="HI_THRESH,LOW_THRESH",
                        help="1,1 means no-op, 0.8,0.9 is a mild filter")
    parser.add_argument('--highpass', type=_float_tuple,
                        default=None, metavar="HI_THRESH,LOW_THRESH",
                        help="0,0 means no-op, 0.1,0.2 is a mild filter")
    parser.add_argument('--resample', type=float, default=1,
                        help="Work with resampled images.")
    # parser.add_argument('--exponent', type=_exponent, default="inf",
    #                     help="Either 'inf' or float. See the docs.")
    parser.add_argument('--iters', type=int, default=1,
                        help="How many iterations to guess the right scale "
                        "and angle")
    parser.add_argument('--extend', type=int, metavar="PIXELS", default=0,
                        help="Extend images by the specified amount of pixels "
                        "before the processing (thus eliminating "
                        "edge effects)")
    parser.add_argument('--order', type=int, default=1,
                        help="Interpolation order (1 = linear, 3 = cubic etc)")
    parser.add_argument('--output', '-o',
                        help="Where to save the transformed image.")
    parser.add_argument(
        '--filter-pcorr', type=int, default=0,
        help="Whether to filter during translation detection. Normally not "
        "needed, but when using low-pass filtering, you may need to increase "
        "filter radius (0 means no filtering, 4 should be enough)")
    parser.add_argument(
        '--print-result', action="store_true", default=False,
        help="We don't print anything unless this option is specified")
    parser.add_argument(
        '--print-format', default="scale: %(scale)f +-%(Dscale)g\n"
        "angle: %(angle)f +-%(Dangle)g\n"
        "shift: %(tx)g, %(ty)g +-%(Dt)g\nSuccess: %(success)g\n", type=outmsg,
        help="Print a string (to stdout) in a given format. A dictionary "
        "containing the 'scale', 'angle', 'tx', 'ty', 'Dscale', 'Dangle', "
        "'Dt' and 'success' keys will be passed for string interpolation")
    parser.add_argument(
        '--tile', action="store_true", default=False, help="If the template "
        "is larger than the image, break the template to pieces of size "
        "similar to image size.")
    parser.add_argument('--version', action="version",
                        version="imreg_dft %s" % ird.__version__,
                        help="Just print version and exit")
    parser.add_argument(
        "--angle", type=_constraints("angle"),
        metavar="MEAN[,STD=0]", default=(0, None),
        help="The mean and standard deviation of the expected angle. ")
    parser.add_argument(
        "--scale", type=_constraints("scale"),
        metavar="MEAN[,STD=0]", default=(1, None),
        help="The mean and standard deviation of the expected scale. ")
    parser.add_argument(
        "--tx", type=_constraints("shift"),
        metavar="MEAN[,STD=0]", default=(0, None),
        help="The mean and standard deviation of the expected X translation. ")
    parser.add_argument(
        "--ty", type=_constraints("shift"),
        metavar="MEAN[,STD=0]", default=(0, None),
        help="The mean and standard deviation of the expected Y translation. ")
    loader.update_parser(parser)
    return parser


def main():
    parser = create_parser()

    args = parser.parse_args()

    loader_stuff = loader.settle_loaders(args, (args.template, args.image))

    # We need tuples in the parser and lists further in the code.
    # So we have to do it like this.
    constraints = dict(
        angle=list(args.angle),
        scale=list(args.scale),
        tx=list(args.tx),
        ty=list(args.ty),
    )

    print_format = args.print_format
    if not args.print_result:
        print_format = None

    opts = dict(
        order=args.order,
        filter_pcorr=args.filter_pcorr,
        extend=args.extend,
        low=args.lowpass,
        high=args.highpass,
        show=args.show,
        print_format=print_format,
        iters=args.iters,
        exponent="inf",
        resample=args.resample,
        tile=args.tile,
        constraints=constraints,
        output=args.output,
    )
    opts.update(loader_stuff)
    run(args.template, args.image, opts)


def filter_images(imgs, low, high):
    # lazy import so no imports before run() is really called
    from imreg_dft import utils

    ret = [utils.imfilter(img, low, high) for img in imgs]
    return ret


def apodize(imgs, radius_ratio):
    # lazy import so no imports before run() is really called
    import numpy as np
    from imreg_dft import utils

    ret = []
    # They might have different shapes...
    for img in imgs:
        shape = img.shape

        bgval = np.median(img)
        bg = np.ones(shape) * bgval

        radius = radius_ratio * min(shape)
        apofield = utils.get_apofield(shape, radius)

        bg *= (1 - apofield)
        # not modifying inplace
        toapp = bg + img * apofield
        ret.append(toapp)
    return ret


def _get_resdict(imgs, opts, tosa=None):
    import numpy as np

    tiledim = None
    if opts["tile"]:
        shapes = np.array([np.array(img.shape) for img in imgs])
        if (shapes[0] / shapes[1]).max() > 1.7:
            tiledim = np.min(shapes, axis=0) * 1.1
            # TODO: Establish a translate region constraint of width tiledim * coef

    if tiledim is not None:
        tiles = ird.utils.decompose(imgs[0], tiledim, 0.35)
        resdicts = []
        shifts = np.empty((len(tiles), 2), float)
        succs = np.empty((len(tiles), 1), float)
        for ii, (tile, pos) in enumerate(tiles):
            try:
                resdict = process_images((tile, imgs[1]), opts, None)
            except ValueError:
                # probably incompatible images due to high scale change, so we
                # just add some harmless stuff here and proceed.
                resdict = dict(tx=0, ty=0, success=0)
            resdicts.append(resdict)
            shifts[ii] = np.array((resdict["ty"], resdict["tx"])) + pos
            succs[ii] = resdict["success"]
            if 0:
                print(ii, succs[ii])
                import pylab as pyl
                pyl.figure(); pyl.imshow(tile)
                pyl.show()
        cog_coords = np.sum(shifts * succs, 0) / np.sum(succs)
        weights = succs / (shifts - cog_coords)
        best = np.argmax(weights)
        best = np.argmax(succs)

        resdict = resdicts[best]

        tosa_offset = np.array(imgs[0].shape)[:2] - np.array(tiledim)[:2] + 0.5
        resdict["tvec"] = shifts[best] - tosa_offset / 2.0
        resdict["ty"], resdict["tx"] = resdict["tvec"]
        # In non-tile cases, tosa is transformed in process_images
        if tosa is not None:
            tosa = ird.transform_img_dict(tosa, resdict)
    else:
        resdict = process_images(imgs, opts, tosa)

    return resdict


def run(template, image, opts):
    # lazy import so no imports before run() is really called
    from imreg_dft import imreg

    fnames = (template, image)
    loaders = opts["loaders"]
    loader_img = loaders[1]
    imgs = [loa.load2reg(fname) for fname, loa in zip(fnames, loaders)]

    tosa = None
    saver = None
    outname = opts["output"]
    if outname is not None:
        tosa = loader_img.get2save()
        saver = loader.LOADERS.get_loader(outname)
        tosa = ird.utils.extend_to_3D(tosa, imgs[0].shape[:3])

    resdict = _get_resdict(imgs, opts, tosa)
    im0, im1, im2 = resdict['unextended']

    if opts["print_format"] is not None:
        msg = opts["print_format"] % resdict
        msg = msg.encode("utf-8")
        msg = msg.decode('unicode_escape')
        sys.stdout.write(msg)

    if outname is not None:
        saver.save(outname, tosa, loader_img)

    if opts["show"]:
        # import ipdb; ipdb.set_trace()
        import pylab as pyl
        fig = pyl.figure()
        imreg.imshow(im0, im1, im2, fig=fig)
        # imreg.imshow(imgs[0], imgs[1], im2, fig=fig)
        pyl.show()


def resample(img, coef):
    from scipy import signal
    ret = img
    for axis in range(2):
        newdim = ret.shape[axis] * coef
        ret = signal.resample(ret, newdim, axis=axis)
    return ret


def process_images(ims, opts, tosa=None):
    # lazy import so no imports before run() is really called
    import numpy as np
    from imreg_dft import utils
    from imreg_dft import imreg

    ims = [utils.extend_by(img, opts["extend"]) for img in ims]
    bigshape = np.array([img.shape for img in ims]).max(0)

    ims = filter_images(ims, opts["low"], opts["high"])
    rcoef = opts["resample"]
    if rcoef != 1:
        ims = [resample(img, rcoef) for img in ims]
        bigshape *= rcoef

    # Make the shape of images the same
    ims = [utils.embed_to(np.zeros(bigshape) + utils.get_borderval(img, 5), img)
           for img in ims]

    resdict = imreg.similarity(
        ims[0], ims[1], opts["iters"], opts["order"], opts["constraints"],
        opts["filter_pcorr"], opts["exponent"])

    im2 = resdict.pop("timg")

    # Seems that the reampling simply scales the translation
    resdict["tvec"] /= rcoef
    ty, tx = resdict["tvec"]
    resdict["tx"] = tx
    resdict["ty"] = ty
    resdict["imgs"] = ims
    tform = resdict

    if tosa is not None:
        tosa[:] = ird.transform_img_dict(tosa, tform)

    if rcoef != 1:
        ims = [resample(img, 1.0 / rcoef) for img in ims]
        im2 = resample(im2, 1.0 / rcoef)
        resdict["Dt"] /= rcoef

    resdict["unextended"] = [utils.unextend_by(img, opts["extend"])
                             for img in ims + [im2]]

    return resdict


if __name__ == "__main__":
    main()



import sys
import os.path
import unittest as ut


tl = ut.defaultTestLoader
tests = tl.discover(os.path.dirname(os.path.abspath(__file__)),
                    r'[a-z]*.py')
suite = ut.TestSuite()

# We are the last discovered test
for test in tests:
    suite.addTest(test)

runner = ut.TextTestRunner(verbosity=2)
ret = runner.run(suite)

retcode = 1
if ret.wasSuccessful():
    retcode = 0
sys.exit(retcode)



# -*- coding: utf-8 -*-

import os
import sys
import setuptools as st

sys.path.append(os.path.join('src'))
import imreg_dft

reqs = open('requirements.txt', 'r').read().strip().splitlines()

longdesc = open(os.path.join('doc', 'description.rst'), 'r').read()

st.setup(
    name="imreg_dft",
    version=imreg_dft.__version__,
    author=u"Matěj Týč",
    author_email="matej.tyc@gmail.com",
    description=("Image registration utility using algorithms based on "
                 "discrete Fourier transform (DFT, FFT)"),
    license="BSD",
    url="https://github.com/matejak/imreg_dft",
    package_dir = {'': 'src'},
    packages = st.find_packages('src'),
    entry_points = {
        'console_scripts': [
           'ird = imreg_dft.cli:main',
        ],
    },
    install_requires=reqs,
    extras_require={
        'plotting':  ["matplotlib>=1.2"],
        'loading images': ["pillow>=2.2"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
    long_description=longdesc,
    zip_safe=True,
)



import os

import scipy as sp
import scipy.misc
import matplotlib.pyplot as plt

import imreg_dft as ird


basedir = os.path.join('..', 'examples')
# the TEMPLATE
im0 = sp.misc.imread(os.path.join(basedir, "sample1.png"), True)
# the image to be transformed
im1 = sp.misc.imread(os.path.join(basedir, "sample2.png"), True)
t0, t1 = ird.translation(im0, im1)
# the Transformed IMaGe.
timg = ird.transform_img(im1, tvec=(t0, t1))
ird.imshow(im0, im1, timg)
plt.show()
print(t0, t1)



# -*- coding: utf-8 -*-
# loader.py

# Copyright (c) 2014-?, Matěj Týč
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
The module contains a layer of functionality that allows
abstract saving and loading of files.

The concept is that there is one loader instance per file loaded.
When we want to save a file, we use a loading loader to provide data to
save and then we instantiate a saving loader (if needed) and save the data.

A loader class inherits from :class:`Loader`.
A singleton class :class:`LoaderSet` is the main public interface
of this module. available as a global variable ``LOADERS``.
It keeps track of all registered loaders and takes care
after them (presents them with options, requests etc.)
Loaders are registered as classes using the decorator :func:`loader`.

Individual loaders absolutely have to implement methods :meth:`Loader._save`
and :meth:`Loader._load2reg`.

This module facilitates integration of its functionality by defining
:func:`update_parser` and :func:`settle_loaders`. While the first one can
add capabilities to a parser (or parser group), the second one updates
``LOADERS`` accordingly while given parsed arguments.

Rough edges (but not rough enough to be worth the trouble):

* You can't force different loaders for image, template and output. If you
  need this, you have to rely on autodetection based on file extension.
* Similarly, there is a problem with loader options --- they are shared among
  all loaders. This is both a bug and a feature though.
* To show the loaders help, you have to satisfy the parser by specifying
  a template and image file strings (they don't have to be real filenames tho).
"""

import sys


def _str2nptype(stri):
    import numpy as np
    msg = ("The string '%s' is supposed to correspond to a "
           "numpy type" % stri)
    try:
        typ = eval("np." + stri)
    except Exception as exc:
        msg += " but it is not the case at all - %s." % exc.message
        raise ValueError(msg)
    typestr = type(typ).__name__
    # We allow mock object for people who know what they are doing.
    if typestr not in ("type", "Mock"):
        msg += " but it is a different animal than 'type': '%s'" % typestr
        raise ValueError(msg)
    return typ


def _str2flat(stri):
    assert stri in "R,G,B,V".split(","), \
        "Flat value has to be one of R, G, B, V, is '%s' instead" % stri
    return stri


def flatten(image, char):
    if image.ndim < 3:
        return image
    char2idx = dict(R=0, G=1, B=2)
    ret = None
    if char == "V":
        ret = image.mean(axis=2)
    elif char in char2idx:
        ret = image[:, :, char2idx[char]]
    else:
        # Shouldn't happen
        assert False, "Unhandled - invalid flat spec '%s'" % char
    return ret


class LoaderSet(object):
    _LOADERS = []
    # singleton-like functionality
    _we = None

    def __init__(self):
        if LoaderSet._we is not None:
            return LoaderSet._we
        loaders = [loader() for loader in LoaderSet._LOADERS]
        self.loader_dict = {}
        for loader in loaders:
            self.loader_dict[loader.name] = loader
        self.loaders = sorted(loaders, key=lambda x: x.priority)
        LoaderSet._we = self

    def _choose_loader(self, fname):
        """
        Use autodetection to select a loader to use.

        Returns:
            Loader instance or None if no loader can be used.
        """
        for loader in self.loaders:
            if loader.guessCanLoad(fname):
                return loader
        # Ouch, no loader available!
        return None

    def get_loader(self, fname, lname=None):
        """
        Try to select a loader. Either we know what we want, or an
        autodetection will take place.
        Exceptions are raised when things go wrong.
        """
        if lname is None:
            ret = self._choose_loader(fname)
            if ret is None:
                msg = ("No loader wanted to load '%s' during autodetection"
                       % fname)
                raise IOError(msg)
        else:
            ret = self._get_loader(lname)
        # Make sure that we don't return the same instance multiple times
        ret = ret.spawn()
        return ret

    def _get_loader(self, lname):
        if lname not in self.loader_dict:
            msg = "No loader named '%s'." % lname
            msg += " Choose one of %s." % self.loader_dict.keys()
            raise KeyError(msg)
        return self.loader_dict(lname)

    def get_loader_names(self):
        """
        What are the names of loaders that we know.
        """
        ret = self.loader_dict.keys()
        return tuple(ret)

    @classmethod
    def add_loader(cls, loader_cls):
        """
        Use this method (at early run-time) to register a loader
        """
        cls._LOADERS.append(loader_cls)

    def print_loader_help(self, lname=None):
        """
        Print info about loaders.
        Either print short summary about all loaders, or focus just on one.
        """
        if lname is None:
            msg = "Available loaders: %s\n" % (self.get_loader_names(),)
            # Lowest priority first - they are usually the most general ones
            for loader in self.loaders[::-1]:
                msg += "\n\t%s: %s\n\tAccepts options: %s\n" % (
                    loader.name, loader.desc, tuple(loader.opts.keys()))
        else:
            loader = self.loader_dict[lname]
            msg = "Loader '%s':\n" % loader.name
            msg += "\t%s\n" % loader.desc
            msg += "Accepts options:\n"
            for opt in loader.opts:
                msg += "\t'%s' (default '%s'): %s\n" % (
                    opt, loader.defaults[opt], loader.opts[opt], )
        print(msg)

    def distribute_opts(self, opts):
        """
        Propagate loader options to all loaders.
        """
        if opts is None:
            # don't return, do something so possible problems surface.
            opts = {}
        for loader in self.loaders:
            loader.setOpts(opts)


def loader(lname, priority):
    """
    A decorator interconnecting an abstract loader with the rest of imreg_dft
    It sets the "nickname" of the loader and its priority during autodetection
    """
    def wrapped(cls):
        cls.name = lname
        cls.priority = priority
        LoaderSet.add_loader(cls)
        return cls
    return wrapped



class Loader(object):
    """

    .. automethod:: _save
    .. automethod:: _load2reg
    """
    name = None
    priority = 10
    desc = ""
    opts = {}
    defaults = {}
    str2val = {}

    def __init__(self):
        self.loaded = None
        self._opts = {}
        # First run, the second will hopefully follow later
        self.setOpts({})
        # We may record some useful stuff for saving during loading
        self.saveopts = {}

    def spawn(self):
        """
        Makes a new instance of the object's class
        BUT it conserves vital data.
        """
        cls = self.__class__
        ret = cls()
        # options passed on command-line
        ret._opts = self._opts
        return ret

    def setOpts(self, options):
        for opt in self.opts:
            stri = options.get(opt, self.defaults[opt])
            val = self.str2val.get(opt, lambda x: x)(stri)
            self._opts[opt] = val

    def guessCanLoad(self, fname):
        """
        Guess whether we can load a filename just according to the name
        (extension)
        """
        return False

    def load2reg(self, fname):
        """
        Given a filename, it loads it and returns in a form suitable for
        registration (i.e. float, flattened, ...).
        """
        try:
            ret = self._load2reg(fname)
        except IOError as err:
            print("Couldn't load '%s': %s" % (fname, err.strerror))
            sys.exit(1)

        return ret

    def get2save(self):
        assert self.loaded is not None, \
            "Saving without loading beforehand, which is not supported. "
        return self.loaded

    def _load2reg(self, fname):
        """
        To be implemented by derived class.
        Load data from fname in a way that they can be used in the
        registration process (so it is a 2D array).
        Possibly take into account options passed upon the class creation.
        """
        raise NotImplementedError("Use the derived class")

    def _save(self, fname):
        """
        To be implemented by derived class.
        Save data to fname, possibly taking into account previous loads
        and/or options passed upon the class creation.
        """
        raise NotImplementedError("Use the derived class")

    def save(self, fname, what, loader):
        """
        Given the registration result, save the transformed input.
        """
        sopts = loader.saveopts
        self.saveopts.update(sopts)
        self._save(fname, what)


@loader("mat", 10)
class _MatLoader(Loader):
    desc = "Loader of .mat (MATLAB v5) binary files"
    opts = {"in": "The structure to load (empty => autodetect)",
            "out": "The structure to save the result to (empty => the same "
                   "as the 'in'",
            "type": "Name of the numpy data type for the output (such as "
                    "int, uint8 etc.)",
            "flat": "How to flatten (the possibly RGB image) for the "
                    "registration. Values can be R, G, B or V (V for value - "
                    "a number proportional to average of R, G and B)",
            }
    defaults = {"in": "", "out": "", "type": "float", "flat": "V"}
    str2val = {"type": _str2nptype, "flat": _str2flat}

    def __init__(self):
        super(_MatLoader, self).__init__()
        # By default, we have not loaded anything
        self.saveopts["loaded_all"] = {}

    def _load2reg(self, fname):
        from scipy import io
        mat = io.loadmat(fname)
        if self._opts["in"] == "":
            valid = [key for key in mat if not key.startswith("_")]
            if len(valid) != 1:
                raise RuntimeError(
                    "You have to supply an input key, there is an ambiguity "
                    "of what to load, candidates are: %s" % (tuple(valid),))
            else:
                key = valid[0]
        else:
            key = self._opts["in"]
            keys = mat.keys()
            if not key in keys:
                raise LookupError(
                    "You requested load of '%s', but you can only choose from"
                    " %s" % (tuple(keys),))
        ret = mat[key]
        self.saveopts["loaded_all"] = mat
        self.saveopts["key"] = key
        self.loaded = ret
        # flattening is a no-op on 2D images
        ret = flatten(ret, self._opts["flat"])
        return ret

    def _save(self, fname, tformed):
        from scipy import io
        if self._opts["out"] == "":
            assert "key" in self.saveopts, \
                "Don't know how to save the output - what .mat struct?"
            key = self.saveopts["key"]
        else:
            key = self._opts["out"]
        out = self.saveopts["loaded_all"]
        out[key] = tformed.astype(self._opts["type"])
        io.savemat(fname, out)

    def guessCanLoad(self, fname):
        return fname.endswith(".mat")


@loader("pil", 50)
class _PILLoader(Loader):
    desc = "Loader of image formats that Pillow (or PIL) can support"
    opts = {"flat": _MatLoader.opts["flat"]}
    defaults = {"flat": _MatLoader.defaults["flat"]}
    str2val = {"flat": _MatLoader.str2val["flat"]}

    def __init__(self):
        super(_PILLoader, self).__init__()

    def _load2reg(self, fname):
        from scipy import misc
        loaded = misc.imread(fname)
        self.loaded = loaded
        ret = loaded
        # flattening is a no-op on 2D images
        ret = flatten(ret, self._opts["flat"])
        return ret

    def _save(self, fname, tformed):
        from scipy import misc
        img = misc.toimage(tformed)
        img.save(fname)

    def guessCanLoad(self, fname):
        "We think that we can do everything"
        return True


@loader("hdr", 10)
class _HDRLoader(Loader):
    desc = ("Loader of .hdr and .img binary files. Supply the '.hdr' as input,"
            "a '.img' with the same basename is expected.")
    opts = {"norm": "Whether to divide the value by 255.0 (0 for not to)"}
    defaults = {"norm": "1"}

    def __init__(self):
        super(_HDRLoader, self).__init__()

    def guessCanLoad(self, fname):
        return fname.endswith(".hdr")

    def _load2reg(self, fname):
        """Return image data from img&hdr uint8 files."""
        import numpy as np
        basename = fname.rstrip(".hdr")
        with open(basename + '.hdr', 'r') as fh:
            hdr = fh.readlines()
        img = np.fromfile(basename + '.img', np.uint8, -1)
        img.shape = int(hdr[4].split()[-1]), int(hdr[3].split()[-1])
        if int(self._opts["norm"]):
            img = img.astype(np.float64)
            img /= 255.0
        return img

    def _save(self, fname, tformed):
        import numpy as np
        # Shouldn't happen, just to make sure
        tformed[tformed > 1.0] = 1.0
        tformed[tformed < 0.0] = 0.0
        tformed *= 255.0
        uint = tformed.astype(np.uint8)
        uint.tofile(fname)


def _parse_opts(stri):
    from argparse import ArgumentTypeError
    components = stri.split(",")
    ret = {}
    for comp in components:
        sides = comp.split("=")
        if len(sides) != 2:
            raise ArgumentTypeError(
                "The options spec has to look like 'option=value', got %s."
                % comp)
        lhs, rhs = sides
        valid_optname = False
        for loader in LOADERS.loaders:
            if lhs in loader.opts:
                valid_optname = True
                break
        if not valid_optname:
            raise ArgumentTypeError(
                "The option '%s' is not understood by any loader" % lhs)
        ret[lhs] = rhs
    return ret


def update_parser(parser):
    parser.add_argument(
        "--loader", choices=LOADERS.get_loader_names(), default=None,
        help="Force usage of a concrete loader (default is autodetection). "
        "If you plan on using two types of loaders to load input, or save the"
        " output, autodetection is the only way to achieve this.")
    parser.add_argument(
        "--loader-opts", default=None, type=_parse_opts,
        help="Options for a loader "
        "(use --loader to make sure that one is used or read the docs.)")
    parser.add_argument(
        "--help-loader", default=False, action="store_true",
        help="Get help on all loaders or on the current loader "
        "and its options.")


def settle_loaders(args, fnames=None):
    if args.help_loader:
        LOADERS.print_loader_help(args.loader)
        sys.exit(0)
    LOADERS.distribute_opts(args.loader_opts)
    ret = {}
    loaders = []
    if fnames is not None:
        for fname in fnames:
            loader = LOADERS.get_loader(fname, args.loader)
            loaders.append(loader)
    ret["loaders"] = loaders
    return ret


LOADERS = LoaderSet()
